#include "reader.hpp"

#include <gflags/gflags.h>

namespace rtransfer {

Reader::Reader(StoragesMap &storages, HandleCache &handleCache,
    std::shared_ptr<ServerSideLink> serverLink)
    : storages_{storages}
    , handleCache_{handleCache}
    , serverLink_{std::move(serverLink)}
{
}

folly::Future<std::size_t> Reader::read(std::uint64_t reqId,
    folly::fbstring storageId, folly::fbstring fileId, folly::fbstring fileGuid,
    std::uint64_t offset, std::size_t size, std::uint8_t priority)
{
    auto it = storages_.find(storageId);
    if (it == storages_.cend())
        return folly::makeFuture<std::size_t>(
            folly::make_exception_wrapper<std::runtime_error>(
                "invalid storage"));

    VLOG(2) << "Reading " << reqId << " offset: " << offset
            << ", size: " << size;
    return handleCache_
        .read(*it->second, reqId, fileId, fileGuid, offset, size, priority)
        .via(folly::getCPUExecutor().get())
        .then([this, reqId, priority, offset](folly::IOBufQueue buf) {
            auto now = std::chrono::steady_clock::now();
            std::size_t size = buf.chainLength();

            auto metric =
                "link." + serverLink_->getPeerAddr() + ".pre_write_buffer.size";
            ONE_METRIC_COUNTER_ADD(metric, size);

            VLOG(2) << "Sending " << reqId << ", offset: " << offset
                    << ", size: " << size;

            auto start = now;
            return folly::collect(sendParts(std::move(buf), reqId, offset, now))
                .via(folly::getCPUExecutor().get())
                .then([=, metric = std::move(metric)](
                          const std::vector<folly::Unit> &) {
                    auto now = std::chrono::steady_clock::now();
                    auto delay = now - start;
                    ONE_METRIC_COUNTER_SUB(metric, size);
                    VLOG(2) << "Sent " << reqId << ", offset: " << offset
                            << ", size: " << size << ", took: "
                            << std::chrono::duration_cast<
                                   std::chrono::milliseconds>(delay)
                                   .count()
                            << "ms";

                    return size;
                });
        });
}
std::vector<folly::Future<folly::Unit>> Reader::sendParts(folly::IOBufQueue buf,
    std::uint64_t reqId, std::uint64_t offset,
    std::chrono::steady_clock::time_point now)
{
    std::size_t size = buf.chainLength();
    std::vector<folly::Future<folly::Unit>> writeFutures;

    auto dataConns = serverLink_->dataConnections(now);
    if (dataConns.empty()) {
        writeFutures.emplace_back(folly::makeFuture<folly::Unit>(
            folly::make_exception_wrapper<std::runtime_error>(
                "no data connections")));
        return writeFutures;
    }

    static thread_local std::mt19937 rng{folly::randomNumberSeed()};
    std::shuffle(dataConns.begin(), dataConns.end(), rng);

    static thread_local folly::fbvector<double> throughput;
    throughput.clear();
    double totalThroughput = 0;
    for (auto &el : dataConns) {
        auto &ti = el.second;
        double tput =
            static_cast<double>(ti.mss * ti.cwnd) / (ti.rtt.count() + 1);
        throughput.push_back(tput);
        totalThroughput += tput;
    }

    static thread_local folly::fbvector<std::size_t> partSize;
    partSize.clear();
    std::size_t assignedSize = 0;
    for (auto tput : throughput) {
        std::size_t s = totalThroughput > 0 ? size * tput / totalThroughput
                                            : size / dataConns.size();
        partSize.push_back(s);
        assignedSize += s;
    }

    // Send any remaining bytes through the first connection
    if (assignedSize < size) {
        partSize.front() += size - assignedSize;
        assignedSize = size;
    }

    for (auto i = 0u, part = 0u; size > 0 && i < dataConns.size(); ++i) {
        if (partSize[i] == 0)
            continue;

        // We use size to check if there's anything left to send in case the
        // floating point calculations somehow lead to assignedSize > size
        size -= std::min(partSize[i], size);

        auto msg = constructDataMsg(
            reqId, offset, part++, size == 0, buf.splitAtMost(partSize[i]));
        writeFutures.emplace_back(dataConns[i].first->write(std::move(msg)));
    }

    return writeFutures;
}

std::unique_ptr<folly::IOBuf> Reader::constructDataMsg(std::uint64_t reqId,
    std::uint64_t offset, std::uint8_t part, bool isLastPart,
    std::unique_ptr<folly::IOBuf> buf)
{
    DCHECK(part < (1 << 7));

    auto nativeSize = buf->computeChainDataLength();

    std::uint32_t size = folly::Endian::big<std::uint32_t>(nativeSize);
    reqId = folly::Endian::big<std::uint64_t>(reqId);
    offset = folly::Endian::big<std::uint64_t>(offset);

    if (isLastPart)
        part |= proto::last_part_mask;

    auto msg = folly::IOBuf::createCombined(
        sizeof(reqId) + sizeof(part) + sizeof(offset) + sizeof(size));

    auto append = [&](auto val) {
        std::memcpy(msg->writableTail(), &val, sizeof(val));
        msg->append(sizeof(val));
    };

    append(reqId);
    append(part);
    append(offset);
    append(size);

    if (nativeSize > 0)
        msg->prependChain(std::move(buf));

    return msg;
}

}  // namespace rtransfer
