#include "storage.hpp"

#include "uuid.hpp"

#include <folly/FBString.h>
#include <folly/Function.h>
#include <folly/String.h>
#include <folly/executors/NamedThreadFactory.h>
#include <folly/executors/ThreadPoolExecutor.h>
#include <folly/io/IOBufQueue.h>
#include <gflags/gflags.h>
#include <helpers/storageHelper.h>
#include <monitoring/monitoring.h>

#include <atomic>
#include <chrono>
#include <queue>
#include <unordered_map>

DEFINE_uint64(
    storage_buckets, 100, "max number of in-progress operations on storage");

namespace {
folly::Future<folly::IOBufQueue> readLoop(one::helpers::FileHandlePtr handle,
    std::uint64_t offset, std::size_t size, folly::IOBufQueue buf)
{
    if (size == 0)
        return folly::makeFuture<folly::IOBufQueue>(std::move(buf));

    return handle->read(offset, size)
        .then([=, buf = std::move(buf)](folly::IOBufQueue nextBuf) mutable {
            if (nextBuf.empty()) {
                size = 0;
            }
            else {
                assert(size >= nextBuf.chainLength());
                size -= nextBuf.chainLength();
                offset += nextBuf.chainLength();
                buf.append(std::move(nextBuf));
            }
            return readLoop(std::move(handle), offset, size, std::move(buf));
        });
}

folly::Future<folly::Unit> writeLoop(one::helpers::FileHandlePtr handle,
    std::uint64_t offset, folly::IOBufQueue buf)
{
    if (buf.empty())
        return folly::makeFuture();

    folly::IOBufQueue cloneBuf{folly::IOBufQueue::cacheChainLength()};
    cloneBuf.append(buf.front()->clone());

    return handle->write(offset, std::move(buf))
        .then([=, cloneBuf = std::move(cloneBuf)](std::size_t wrote) mutable {
            cloneBuf.split(wrote);
            return writeLoop(std::move(handle), offset, std::move(cloneBuf));
        });
}
}  // namespace

namespace rtransfer {

Storage::Storage(
    folly::fbstring storageId, one::helpers::StorageHelperPtr storageHelper)
    : storageIdHex_{folly::hexlify(storageId)}
    , storageHelper_{std::move(storageHelper)}
    , readQueue_{FLAGS_storage_buckets}
    , writeQueue_{FLAGS_storage_buckets}
{
    LOG(INFO) << "Adding storage " << storageIdHex_;
}

folly::Future<one::helpers::FileHandlePtr> Storage::open(
    folly::fbstring fileId, OpenType openType)
{
    VLOG(2) << "Opening file '" << fileId << "'";
    const int openFlag = openType == OpenType::read ? O_RDONLY : O_WRONLY;
    return storageHelper_->open(fileId, openFlag, {});
}

folly::Future<folly::IOBufQueue> Storage::read(
    one::helpers::FileHandlePtr handle, std::uint64_t reqId,
    const std::uint64_t offset, const std::size_t size,
    const std::uint8_t priority)
{
    ReadRequest req{reqId, {}, std::move(handle), offset, size};
    auto future = req.promise.getFuture();
    executor_.add([=, req = std::move(req)]() mutable {
        auto size = req.size;
        auto bucket = readQueue_.bucket(req);
        VLOG(2) << "Pushing " << reqId << " into readQueue bucket " << bucket;
        readQueue_.get(bucket).push(std::move(req));
        if (!readQueue_.isUsed(bucket))
            handleRead(bucket);
        else {
            ONE_METRIC_COUNTER_INC(m(".reads.queue.items"));
            ONE_METRIC_COUNTER_ADD(m(".reads.queue.size"), size);
        }
    });
    return future;
}

std::string Storage::m(const char *name)
{
    return "storage." + storageIdHex_.toStdString() + name;
}

folly::Future<folly::Unit> Storage::write(one::helpers::FileHandlePtr handle,
    std::uint64_t reqId, const std::uint64_t offset, folly::IOBufQueue buf,
    const std::uint8_t priority)
{
    WriteRequest req{reqId, {}, std::move(handle), offset, std::move(buf)};
    auto future = req.promise.getFuture();
    executor_.add([=, req = std::move(req)]() mutable {
        auto size = req.buf.chainLength();
        auto bucket = writeQueue_.bucket(req);
        VLOG(2) << "Pushing " << reqId << " into writeQueue bucket " << bucket;
        writeQueue_.get(bucket).push(std::move(req));
        if (!writeQueue_.isUsed(bucket))
            handleWrite(bucket);
        else {
            ONE_METRIC_COUNTER_INC(m(".writes.queue.items"));
            ONE_METRIC_COUNTER_ADD(m(".writes.queue.size"), size);
        }
    });
    return future;
}

template <typename Queue, typename Cb>
void Storage::usingBucket(Queue &queue, const std::size_t bucket, Cb &&cb)
{
    queue.use(bucket);
    auto reqs = queue.get(bucket).pop();
    cb(std::move(reqs))
        .via(&executor_)
        .ensure([=, &queue, cb = std::forward<Cb>(cb)]() mutable {
            if (!queue.get(bucket).empty())
                usingBucket(queue, bucket, std::move(cb));
            else
                queue.stopUsing(bucket);
        });
}

template <typename T>
void logRequests(const char *what, const folly::fbvector<T> &reqs)
{
    if (!VLOG_IS_ON(2))
        return;

    std::stringstream log;
    for (auto &req : reqs)
        log << " {reqId: " << req.reqId << ", offset: " << req.offset
            << ", size: " << req.getSize() << "}";
    VLOG(2) << what << " requests:" << log.str();
}

void Storage::handleRead(const std::size_t bucket)
{
    usingBucket(readQueue_, bucket, [this](folly::fbvector<ReadRequest> reqs) {
        logRequests("Reading", reqs);
        auto handle = reqs.front().handle;
        auto offset = reqs.front().offset;
        std::size_t size = 0;
        for (auto &req : reqs)
            size += req.size;

        return readLoop(handle, offset, size,
            folly::IOBufQueue{folly::IOBufQueue::cacheChainLength()})
            .then([reqs = std::move(reqs)](
                      folly::Try<folly::IOBufQueue> buft) mutable {
                if (buft.hasValue()) {
                    logRequests("Read", reqs);
                    auto buf = std::move(buft.value());
                    for (auto &req : reqs) {
                        folly::IOBufQueue subQueue{
                            folly::IOBufQueue::cacheChainLength()};
                        subQueue.append(buf.split(req.size));
                        req.promise.setValue(std::move(subQueue));
                    }
                }
                else {
                    logRequests("Failed to read", reqs);
                    for (auto &req : reqs)
                        req.promise.setException(buft.exception());
                }
            });
    });
}

void Storage::handleWrite(const std::size_t bucket)
{
    usingBucket(
        writeQueue_, bucket, [this](folly::fbvector<WriteRequest> reqs) {
            logRequests("Writing", reqs);
            auto handle = reqs.front().handle;
            auto offset = reqs.front().offset;
            folly::IOBufQueue buf{folly::IOBufQueue::cacheChainLength()};
            for (auto &req : reqs)
                buf.append(std::move(req.buf));

            return writeLoop(handle, offset, std::move(buf))
                .then([reqs = std::move(reqs)](
                          folly::Try<folly::Unit> t) mutable {
                    if (t.hasValue()) {
                        logRequests("Wrote", reqs);
                        for (auto &req : reqs)
                            req.promise.setValue();
                    }
                    else {
                        logRequests("Failed to write", reqs);
                        for (auto &req : reqs)
                            req.promise.setException(t.exception());
                    }
                });
        });
}

}  // namespace rtransfer
